using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel.Channels;
using System.ServiceModel.Dynamic.Test.ServiceModel.Cache;
using System.ServiceModel.Dynamic.Test.ServiceModel.Cache.Configuration;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache
{
    /// <summary>
    /// A binding provides a factory that knows how to create the channel that executes at runtime (phew!)
    /// </summary>
    public class ClientCacheBindingElement : BindingElement
    {

        #region Private Fields
        private TimeSpan timeout;
        private KeyCreationMethod keyCreationMethod;
        private string cacheType;
        private bool enabled;
        private bool header;
        private List<int> indexList = new List<int>();
        private Dictionary<string, OperationSettings> operationDictionary;
        #endregion

        #region Public Constructors

        public ClientCacheBindingElement(TimeSpan timeout,                                         
                                         string indexes,
                                         KeyCreationMethod keyCreationMethod,
                                         string cacheType,
                                         bool enabled,
                                         bool header,
                                         OperationElementCollection operations)
        {

            this.timeout = timeout;
            this.enabled = enabled;
            this.header = header;

            if (!System.ServiceModel.Dynamic.Test.ServiceModel.Cache.CacheFactory.Instance.Exists(this.cacheType))
            {
                // Sets the default value specified at the root level.
                this.cacheType = System.ServiceModel.Dynamic.Test.ServiceModel.Cache.CacheFactory.DefaultCache;
            }

            if (!string.IsNullOrEmpty(indexes))
            {
                int index;
                string[] items = indexes.Split(',');
                for (int i = 0; i < items.Length; i++)
                {
                    if (int.TryParse(items[i], out index) &&
                        !indexList.Contains(index))
                    {
                        indexList.Add(index);
                    }
                }
            }

            this.operationDictionary = new Dictionary<string, OperationSettings>();

            OperationSettings operationSettings;

            for (int i = 0; i < operations.Count; i++)
            {
                operationSettings = new OperationSettings(operations[i].Indexes);

                operationSettings.KeyCreationMethod = operations[i].KeyCreationMethod;

                if (!operationDictionary.ContainsKey(operations[i].Action))
                {
                    // The following table shows the result of the interaction for the value 
                    // of the Enabled property the root and action level.
                    //
                    //               Action 
                    //          | True  | False |
                    // R        -----------------
                    // o  True  | True  | False |
                    // o  -----------------------
                    // t  False | False | False |
                    //    -----------------------
                    operationSettings.Enabled = this.enabled && operations[i].Enabled;

                    // The following table shows the result of the interaction for the value 
                    // of the Header property the root and action level.
                    //
                    //               Action 
                    //          | True  | False |
                    // R        -----------------
                    // o  True  | True  | True  |
                    // o  -----------------------
                    // t  False | True  | False |
                    //    -----------------------
                    operationSettings.Header = this.enabled ? true : operations[i].Header;

                    if (!CacheFactory.Instance.Exists(operationSettings.CacheType))
                    {
                        operationSettings.CacheType = CacheFactory.DefaultCache;
                    }

                    if (operations[i].Timeout.TotalSeconds >= 10)
                    {
                        operationSettings.Timeout = operations[i].Timeout;
                    }
                    else
                    {
                        operations[i].Timeout = this.timeout;
                    }
                    operationDictionary.Add(operations[i].Action, operationSettings);
                }
            }
        }

        public ClientCacheBindingElement(ClientCacheBindingElement other) : base(other)
        {
            this.timeout = other.timeout;
            this.keyCreationMethod = other.keyCreationMethod;
            this.cacheType = other.cacheType;
            this.operationDictionary = other.operationDictionary;
        }
        #endregion

        #region Public Properties
        /// <summary>
        /// Gets or sets the amount of time the object should reside in the cache before expiration.
        /// </summary>
        public TimeSpan Timeout
        {
            get 
            { 
                return timeout; 
            }
            set 
            { 
                timeout = value; 
            }
        }


        /// <summary>
        /// Gets or sets the method used to calculate the key of the cache item.
        /// </summary>
        public KeyCreationMethod KeyCreationMethod
        {
            get
            {
                return this.keyCreationMethod;
            }
            set
            {
                this.keyCreationMethod = value;
            }
        }

        /// <summary>
        /// Gets or sets the cache type used to store items.
        /// </summary>
        public string CacheType
        {
            get
            {
                return this.cacheType;
            }
            set
            {
                this.cacheType = value;
            }
        } 
        #endregion

        #region Public Methods

        public override BindingElement Clone()
        {
            return new ClientCacheBindingElement(this);
        }

        public override T GetProperty<T>(BindingContext context)
        {
            return context.GetInnerProperty<T>();
        }

        public override bool CanBuildChannelFactory<TChannel>(BindingContext context)
        {
            return context.CanBuildInnerChannelFactory<TChannel>();
        }

        public override bool CanBuildChannelListener<TChannel>(BindingContext context)
        {
            return false;
        }

        public override IChannelFactory<TChannel> BuildChannelFactory<TChannel>(BindingContext context)
        {

            return new ClientCacheChannelFactory<TChannel>(context, 
                                                           timeout, 
                                                           keyCreationMethod, 
                                                           cacheType,
                                                           enabled,
                                                           header,
                                                           indexList,
                                                           operationDictionary);

        }

        public override IChannelListener<TChannel> BuildChannelListener<TChannel>(BindingContext context)
        {
            throw new NotImplementedException();
        } 

        #endregion
    
    }
}
