using System;
using System.IO;
using System.Diagnostics;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.XPath;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Dynamic.Test.ServiceModel.Cache;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache
{

	public class ClientCacheRequestChannel<TChannel> : ClientCacheChannelBase<IRequestChannel>, IRequestChannel, System.ServiceModel.Channels.IRequestSessionChannel
	{

		#region Private Constants
		private const string DefaultSettingsKey = "DefaultSettingsKey";
		private const string headerName = "source";
		private const string headerNamespace = "http://appfabric.cat.microsoft.com/10/samples/wcf/clientcaching";
		private const string cacheSource = "Cache";
		private const string serviceSource = "Service";
		#endregion

		#region Private Fields
		private TimeSpan timeout;
		private KeyCreationMethod keyCreationMethod;
		private string cacheType;
		private bool enabled;
		private bool header;
		private List<int> indexList;
		private Dictionary<string, OperationSettings> operationDictionary;
		#endregion

		#region Public Constructors
		public ClientCacheRequestChannel(ClientCacheChannelFactory<TChannel> factory, 
										 IRequestChannel innerChannel, 
										 TimeSpan timeout, 										 
										 KeyCreationMethod keyCreationMethod,
										 string cacheType,
										 bool enabled,
										 bool header,
										 List<int> indexList,
										 Dictionary<string, OperationSettings> operationDictionary)
			: base(factory, innerChannel)
		{
			this.timeout = timeout;
			this.keyCreationMethod = keyCreationMethod;
			this.cacheType = cacheType;
			this.enabled = enabled;
			this.header = header;
			this.indexList = indexList;
			this.operationDictionary = operationDictionary;
			if (!this.operationDictionary.ContainsKey(DefaultSettingsKey))
			{
				this.operationDictionary.Add(DefaultSettingsKey, new OperationSettings(indexList, keyCreationMethod, cacheType, timeout, enabled, header));
			}
		} 
		#endregion

		#region IRequestChannel Members

		public IAsyncResult BeginRequest(Message message, AsyncCallback callback, object state)
		{
			return this.BeginRequest(message, this.DefaultSendTimeout, callback, state);
		}

		public IAsyncResult BeginRequest(Message message, TimeSpan timeout, AsyncCallback callback, object state)
		{
			Message reply = this.Request(message);
			return new TypedCompletedAsyncResult<Message>(reply, callback, state);
		}

		public Message EndRequest(IAsyncResult result)
		{
			TypedCompletedAsyncResult<Message> reply = (TypedCompletedAsyncResult<Message>)result;
			return reply.Data;
		}

		public System.ServiceModel.EndpointAddress RemoteAddress
		{
			get { return this.InnerChannel.RemoteAddress; }
		}

		public Message Request(Message request, TimeSpan timeout)
		{
			// This variable will contain the response to deliver back, from cache or inner channel
			Message response; 
			// Store the cacheKey evaluated for request.
			// This will eventually be used to store the response in the cache 
			// should the custom channel call the target service via its inner channel
			string cacheKey;
			// Retrieve operation settings from the dictionary
			OperationSettings operationSettings = GetOperationSettings(request.Headers.Action);
			if (operationSettings.Enabled)
			{
				// Try to get the response from the cache
				response = ReadFromCache(operationSettings, ref request, out cacheKey);
				if (response == null)
				{
					response = this.InnerChannel.Request(request);
					if (operationSettings.Header)
					{
						response.Headers.Add(MessageHeader.CreateHeader(headerName, headerNamespace, serviceSource));
					}
					// The message is not cached if it's a fault or empty.
					if (response != null &&
						!response.IsFault &&
						!response.IsEmpty)
					{
						WriteToCache(operationSettings, cacheKey, ref response);
					}
				}
			}
			else
			{
				response = this.InnerChannel.Request(request);
				if (operationSettings.Header)
				{
					response.Headers.Add(MessageHeader.CreateHeader(headerName, headerNamespace, serviceSource));
				}
			}
			return response;
		}

		public Message Request(Message message)
		{
			return this.Request(message, this.DefaultSendTimeout);
		}

		public Uri Via
		{
			get 
			{ 
				return this.InnerChannel.Via; 
			}
		}

		#endregion

		#region Private Methods
		/// <summary>
		/// Returns an instance of the OperationInfo class which contains settings for the call.
		/// </summary>
		/// <param name="action">The WS-Addressing action of the request message.</param>
		/// <returns></returns>
		private OperationSettings GetOperationSettings(string action)
		{


			string actionName = action;

			if (operationDictionary.Count > 0)
			{
				if (actionName.Contains("/") && !operationDictionary.First().Key.Contains("/"))
				{
					string[] tb = actionName.Split('/');
					actionName = tb[tb.Length - 1];
				}
			}
			if (operationDictionary.ContainsKey(actionName))
				return operationDictionary[actionName];
			else
				return operationDictionary[DefaultSettingsKey];
		}

		/// <summary>
		/// Attempts to find a response for the request provided in the cache
		/// Returns a response if found, or null otherwise.
		/// Also provides the key used for the cache should the caller want to update the cache for this request
		/// The request message will be re-created if the response was not found in the cache so that the caller can send it to the inner channel
		/// </summary>
		/// <param name="request">The operation settings.</param>
		/// <param name="request">The request for which a response is needed, used to extract the cache key to use.</param>
		/// <param name="cacheKey">The cache key evaluated for the request.</param>
		/// <returns>A response message if found in the cache, null otherwise.</returns>
		private Message ReadFromCache(OperationSettings operationSettings, ref Message request, out string cacheKey)
		{
			cacheKey = null;
			MessageBuffer requestBuffer = null;
			try 
			{
				if (request != null)
				{
					// Evaluate the cache key for the request, also returns the buffer of the request message (as it is being read by the function)
                    requestBuffer = request.CreateBufferedCopy(request.ToString().Length);
					cacheKey = KeyCreationHelper.GenerateKey(requestBuffer.CreateMessage(), operationSettings);
					Message cachedResponse = null;
					// Attempt to get a response from the cache as a message buffer
					MessageCacheItem messageCacheItem = null;

					ICache c = operationSettings.CacheIstance;
					messageCacheItem = c.Get<MessageCacheItem>(cacheKey);

					if (messageCacheItem != null)
					{
						MemoryStream messageCopy = new MemoryStream(messageCacheItem.Buffer);
						messageCopy.Seek(0, SeekOrigin.Begin);
						cachedResponse = Message.CreateMessage(request.Version,
															   messageCacheItem.Action,
															   new XmlTextReader(messageCopy));
						if (operationSettings.Header)
						{
							cachedResponse.Headers.Add(MessageHeader.CreateHeader(headerName, headerNamespace, cacheSource));
						}
					}
					if (cachedResponse == null)
					{
						// If we haven't found a cached response the request will need to be sent through the underlying channel.
						// as we've read it we have to create a new message from the buffer.
						request = requestBuffer.CreateMessage();
						return null;
					}
					else
					{
						// Return the cached response

						return cachedResponse;
					}
				}
			}
			finally
			{
				// Close the request buffer
				if (requestBuffer != null)
				{
					requestBuffer.Close();
				}
			}
			return null;
		}

		/// <summary>
		/// Writes a message in the cache.
		/// </summary>
		/// <param name="request">The operation settings.</param>
		/// <param name="cacheKey">The unique key that is used to identify the object in the cache.</param>
		/// <param name="response">The message to add  to the cache.</param>
		private void WriteToCache(OperationSettings operationSettings, string cacheKey, ref Message response)
		{
			MessageBuffer responseBuffer = null;

            try
            {

                responseBuffer = null;
                int lenght = response.ToString().Length;
                responseBuffer = response.CreateBufferedCopy(lenght);

               

                // Store the buffer in the cache using the key provided (which was evaluated from the corresponding request)
                // Create a message cache item
                // Note: neither the Message class nor the MessageBuffer class are serializable with 
                // DataContractSerializer. Therefore, I chose to create the MessageCacheItem class
                // to save the data necessary to rebuild the response message.
                MemoryStream stream = new MemoryStream(responseBuffer.BufferSize);
                Message messageCopy = responseBuffer.CreateMessage();
                XmlWriterSettings settings = new XmlWriterSettings();
                settings.OmitXmlDeclaration = true;
                
                using (XmlWriter xmlWriter = XmlWriter.Create(stream, settings))
                {
                    using (XmlDictionaryWriter xmlDictionaryWriter = XmlDictionaryWriter.CreateDictionaryWriter(xmlWriter))
                    {                        
                        messageCopy.WriteBodyContents(xmlDictionaryWriter);
                        xmlDictionaryWriter.Flush();
                        stream.Seek(0, SeekOrigin.Begin);
                    }
                }
                stream.Flush();


                ICache c = operationSettings.CacheIstance;

                TimerExpiration timer = new TimerExpiration(timeout);

                c.Add<MessageCacheItem>(cacheKey, new MessageCacheItem(response.Headers.Action, stream.ToArray(), timer), timer);

                // Create a message out of the buffer and returns as the response
                response = responseBuffer.CreateMessage();
            }            
			finally
			{
				// Close the response buffer
				if (responseBuffer != null)
				{
					responseBuffer.Close();
				}
			}
		}
		
		#endregion

		public IOutputSession Session
		{
			get { return _session; }
		}


		/// <summary>
		/// Internal implementation of a session, with tracking ID.
		/// </summary>
		private class OutputSession : IOutputSession, System.ServiceModel.Channels.ISession
		{
			private string _sessionId = NmsChannelHelper.CreateUniqueSessionId();

			/// <summary>
			/// Gets the ID that uniquely identifies the session.
			/// </summary>
			/// <value></value>
			/// <returns>The ID that uniquely identifies the session. </returns>
			public string Id
			{
				get { return _sessionId; }
			}
		}


		#region Private members

		private IOutputSession _session = new OutputSession();

		#endregion

	}



	/// <summary>
	/// A helper class for the NMS transport.
	/// </summary>
	internal class NmsChannelHelper
	{
		/// <summary>
		/// Ensures that the specified timeout value is not negative.
		/// </summary>
		/// <param name="timeout">The timeout that needs to be validated.</param>
		/// <exception cref="ArgumentOutOfRangeException">the timeout value was negative.</exception>
		public static void ValidateTimeout(TimeSpan timeout)
		{
			if (timeout < TimeSpan.Zero)
			{
				throw new ArgumentOutOfRangeException("timeout", timeout, "Timeout must be greater than or equal to TimeSpan.Zero. To disable timeout, specify TimeSpan.MaxValue.");
			}
		}

		/// <summary>
		/// Creates a unique session identifier.
		/// </summary>
		public static string CreateUniqueSessionId()
		{
			return "uuid:/session-gram/" + Guid.NewGuid();
		}
	}

}
