using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel.Configuration;
using System.Configuration;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache.Configuration
{
    /// <summary>
    /// Providers the configuration file support for the ClientCache custom binding element (channel)
    /// </summary>
    public class ClientCacheBindingExtensionElement : BindingElementExtensionElement
    {
        #region Private Constants
        private const string EnabledProperty = "enabled";
        private const string HeaderProperty = "header";
        private const string TimeoutProperty = "timeout";
        private const string IndexesProperty = "indexes";
        private const string CacheNameProperty = "cacheName";
        private const string RegionNameProperty = "regionName";
        private const string OperationsProperty = "operations";
        private const string OperationProperty = "operation";
        private const string KeyCreationMethodProperty = "keyCreationMethod";
        private const string KeyCreationMethodDefaultValue = "MessageBody";
        private const string KeyCreationMethodList = "(Action|MessageBody|Simple|Indexed|MD5)";
        private const string CacheTypeProperty = "cacheType";
        private const string CacheTypeDefaultValue = "MemoryCache";
        private const string CacheTypeList = "(WebCache|MemoryCache|AppFabricCache)";
        private const string FiveMinutes = "00:05:00"; 
        private const string ThirtySeconds = "00:00:30";
        private const string ThirtyDays = "720:00:00";
        private const string CacheNameDefaultValue = "default";
        #endregion

        #region Private Fields
        private ClientCacheBindingElement clientCacheBindingElement;
        #endregion

        #region Public Constructors
        public ClientCacheBindingExtensionElement()
        {
        } 
        #endregion

        #region Public Properties
        /// <summary>
        /// Gets or sets a value indicating whether the WCF client cache is enabled.
        /// When the value is false, the channel stack always invokes the target service.
        /// </summary>
        [ConfigurationProperty(EnabledProperty, DefaultValue = true, IsRequired = false)]
        public bool Enabled
        {
            get
            {
                return (bool)base[EnabledProperty];
            }
            set
            {
                base[EnabledProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether a custom header is added to the response to 
        /// indicate the source of the WCF message (cache or service).
        /// </summary>
        [ConfigurationProperty(HeaderProperty, DefaultValue = false, IsRequired = false)]
        public bool Header
        {
            get
            {
                return (bool)base[HeaderProperty];
            }
            set
            {
                base[HeaderProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets the amount of time the object should reside in the cache before expiration.
        /// </summary>
        [ConfigurationProperty(TimeoutProperty, DefaultValue = FiveMinutes, IsRequired = false)]
        [TimeSpanValidator(MinValueString = ThirtySeconds, MaxValueString = ThirtyDays, ExcludeRange = false)]
        public TimeSpan Timeout
        {
            get
            {
                return (TimeSpan)base[TimeoutProperty];
            }
            set
            {
                base[TimeoutProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets the named cache to use for cache operations.
        /// This property is used only when the CacheType = AppFabricCache.
        /// </summary>
        [ConfigurationProperty(CacheNameProperty, DefaultValue = CacheNameDefaultValue, IsRequired = false)]
        [StringValidator(MinLength = 0, MaxLength = 1024)]
        public string CacheName
        {
            get
            {
                return (string)this[CacheNameProperty];
            }
            set
            {
                this[CacheNameProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets the named region to use for cache operations.
        /// This property is used only when the CacheType = AppFabricCache.
        /// </summary>
        [ConfigurationProperty(RegionNameProperty, DefaultValue = null, IsRequired = false)]
        [StringValidator(MinLength = 0, MaxLength = 1024)]
        public string RegionName
        {
            get
            {
                return (string)this[RegionNameProperty];
            }
            set
            {
                this[RegionNameProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets a string containing a comma-separated list of indexes of parameters to be used to calculate the cache key.
        /// This property is used only when the KeyConstructionMethod = Indexed
        /// </summary>
        [ConfigurationProperty(IndexesProperty, DefaultValue = null, IsRequired = false)]
        [StringValidator(MinLength = 0, MaxLength = 1024)]
        public string Indexes
        {
            get
            {
                return (string)this[IndexesProperty];
            }
            set
            {
                this[IndexesProperty] = value;
            }
        }

        

        /// <summary>
        /// Gets or sets the method used to calculate the key of the cache item.
        /// </summary>
        [ConfigurationProperty(KeyCreationMethodProperty, DefaultValue = KeyCreationMethodDefaultValue, IsRequired = true)]
        [RegexStringValidator(KeyCreationMethodList)]
        public KeyCreationMethod KeyCreationMethod
        {
            get
            {
                return (KeyCreationMethod)this[KeyCreationMethodProperty];
            }
            set
            {
                this[KeyCreationMethodProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets the cache type used to store items.
        /// </summary>
        [ConfigurationProperty(CacheTypeProperty, DefaultValue = CacheTypeDefaultValue, IsRequired = false)]
        [RegexStringValidator(CacheTypeList)]
        public string CacheType
        {
            get
            {
                return (string)this[CacheTypeProperty];
            }
            set
            {
                this[CacheTypeProperty] = value;
            }
        } 

        /// <summary>
        /// Gets a collection containing settings for individual operations
        /// </summary>
        [ConfigurationProperty(OperationsProperty, IsDefaultCollection = true)]
        [ConfigurationCollection(typeof(OperationElementCollection), AddItemName = OperationProperty)]
        public OperationElementCollection Operations
        {
            get
            {
                return (OperationElementCollection)base[OperationsProperty];
            }
        }
        #endregion

        #region Public Methods
        public override Type BindingElementType
        {
            get 
            { 
                return typeof(ClientCacheBindingElement); 
            }
        }

        public override void ApplyConfiguration(System.ServiceModel.Channels.BindingElement bindingElement)
        {
            base.ApplyConfiguration(bindingElement);
            clientCacheBindingElement = (ClientCacheBindingElement)bindingElement;
        }
        #endregion

        #region Protected Methods
        protected override System.ServiceModel.Channels.BindingElement CreateBindingElement()
        {
            //CacheSettings.CacheName = this.CacheName;
            //CacheSettings.RegionName = this.RegionName;
            clientCacheBindingElement = new ClientCacheBindingElement(this.Timeout, 
                                                                      this.Indexes,
                                                                      this.KeyCreationMethod, 
                                                                      this.CacheType, 
                                                                      this.Enabled,
                                                                      this.Header,
                                                                      this.Operations);
            this.ApplyConfiguration(clientCacheBindingElement);
            return clientCacheBindingElement;
        } 
        #endregion
    }
}
