using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel.Configuration;
using System.Configuration;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache.Configuration
{

    public class OperationElement : ConfigurationElement
    {

        #region Private Constants
        private const string EnabledProperty = "enabled";
        private const string HeaderProperty = "header";
        private const string ActionProperty = "action";
        private const string IndexesProperty = "indexes";
        private const string KeyCreationMethodProperty = "keyCreationMethod";
        private const string KeyCreationMethodList = "(Action|MessageBody|Simple|MD5|Indexed|Unknown)";
        private const string CacheTypeProperty = "cacheType";
        private const string CacheTypeList = "(WebCache|MemoryCache|AppFabricCache|Unknown)";
        private const string TimeoutProperty = "timeout";
        private const string FiveSeconds = "00:00:05";
        private const string ThirtyDays = "720:00:00";
        private const string Unknown = "Unknown";
        #endregion

        #region Public Constructors
        public OperationElement()
        {
            this.Action = Unknown;
            this.KeyCreationMethod = KeyCreationMethod.Action;
            this.CacheType = System.ServiceModel.Dynamic.Test.ServiceModel.Cache.CacheFactory.DefaultCache;
            this.Timeout = TimeSpan.FromMinutes(1);
        }


        public OperationElement(string action,
                                KeyCreationMethod keyCreationMethod,
                                string cacheType,
                                TimeSpan timeout,
                                bool enabled)
        {
            this.Action = action;
            this.KeyCreationMethod = keyCreationMethod;
            this.CacheType = cacheType;
            this.Timeout = timeout;
            this.Enabled = enabled;
        }


        #endregion

        #region Public Properties
        /// <summary>
        /// Gets or sets the WS-Addressing action of the request message.
        /// </summary>
        [ConfigurationProperty(ActionProperty, DefaultValue = Unknown, IsRequired = true, IsKey = true)]
        [StringValidator(MinLength=1, MaxLength=1024)]
        public string Action
        {
            get
            {
                return (string)this[ActionProperty];
            }
            set
            {
                this[ActionProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the WCF client cache is enabled.
        /// When the value is false, the channel stack always invokes the target service.
        /// </summary>
        [ConfigurationProperty(EnabledProperty, DefaultValue = true, IsRequired = false)]
        public bool Enabled
        {
            get
            {
                return (bool)base[EnabledProperty];
            }
            set
            {
                base[EnabledProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether a custom header is added to the response to 
        /// indicate the source of the WCF message (cache or service).
        /// </summary>
        [ConfigurationProperty(HeaderProperty, DefaultValue = false, IsRequired = false)]
        public bool Header
        {
            get
            {
                return (bool)base[HeaderProperty];
            }
            set
            {
                base[HeaderProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets a string containing a comma-separated list of indexes of parameters to be used to calculate the cache key.
        /// This property is used only when the KeyConstructionMethod = Indexed
        /// </summary>
        [ConfigurationProperty(IndexesProperty, DefaultValue = null, IsRequired = false)]
        [StringValidator(MinLength = 0, MaxLength = 1024)]
        public string Indexes
        {
            get
            {
                return (string)this[IndexesProperty];
            }
            set
            {
                this[IndexesProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets the method used to calculate the key of the cache item.
        /// </summary>
        [ConfigurationProperty(KeyCreationMethodProperty, DefaultValue = "Simple", IsRequired = false)]
        public KeyCreationMethod KeyCreationMethod
        {
            get
            {
                return (KeyCreationMethod)this[KeyCreationMethodProperty];
            }
            set
            {
                this[KeyCreationMethodProperty] = value;
            }
        }

        /// <summary>
        /// Gets or sets the cache type used to store items.
        /// </summary>
        [ConfigurationProperty(CacheTypeProperty, DefaultValue = Unknown, IsRequired = false)]
        [RegexStringValidator(CacheTypeList)]
        public string CacheType
        {
            get
            {
                return (string)this[CacheTypeProperty];
            }
            set
            {
                this[CacheTypeProperty] = value;
            }
        } 

        /// <summary>
        /// Gets or sets the amount of time the object should reside in the cache before expiration.
        /// </summary>
        [ConfigurationProperty(TimeoutProperty, DefaultValue = FiveSeconds, IsRequired = false)]
        //[TimeSpanValidator(MinValueString = FiveSeconds, MaxValueString = ThirtyDays, ExcludeRange = false)]
        public TimeSpan Timeout
        {
            get
            {
                return (TimeSpan)base[TimeoutProperty];
            }
            set
            {
                base[TimeoutProperty] = value;
            }
        }

        
        #endregion
    
    }
}
