using System;
using System.Xml;
using System.Xml.XPath;
using System.Text;
using System.Diagnostics;
using System.Security.Cryptography;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.Runtime.Serialization;

namespace System.ServiceModel.Dynamic.Test.ServiceModel.Cache
{
    public class KeyCreationHelper
    {
        #region Private Constants
        private const string MessageParameterCannotBeNull = "The message parameter cannot be null";
        private const string WCFClientCachePerformanceCounterCategoryName = "WCF Client Cache";
        private const string WCFClientCachePerformanceCounterCategoryDescription = "This category contains performance counters exposed by the WCF Client Cache component.";
        private const string KeyCreationLatencyPerformanceCounterName = "Key Creation Latency (millisecond)";
        private const string KeyCreationLatencyPerformanceCounterDescription = "Time in milliseconds spent to create a cache key with the selected method.";
        #endregion

        #region Static Fields
        //indicates if performance counters set up on server and so should be used.
        private static bool performanceCountersExist = false;
        private static PerformanceCounter performanceCounterKeyCreationLatency = null;
        #endregion

        #region Static Constructor
        /// <summary>
        /// static constructor used to initialize performance counters
        /// </summary>
        static KeyCreationHelper()
        {
            #region initialise counters
            try
            {
                if (!PerformanceCounterCategory.Exists(WCFClientCachePerformanceCounterCategoryName))
                {
                    // Create a collection of type CounterCreationDataCollection.
                    CounterCreationDataCollection counterCreationDataCollection = new CounterCreationDataCollection();

                    // Create the counters and set their properties.
                    CounterCreationData counterCreationData = new System.Diagnostics.CounterCreationData();
                    counterCreationData.CounterName = KeyCreationLatencyPerformanceCounterName;
                    counterCreationData.CounterHelp = KeyCreationLatencyPerformanceCounterDescription;
                    counterCreationData.CounterType = PerformanceCounterType.NumberOfItems64;

                    // Add both counters to the collection.
                    counterCreationDataCollection.Add(counterCreationData);

                    // Create the category and pass the collection to it.
                    PerformanceCounterCategory.Create(WCFClientCachePerformanceCounterCategoryName,
                                                      WCFClientCachePerformanceCounterCategoryDescription,
                                                      PerformanceCounterCategoryType.MultiInstance,
                                                      counterCreationDataCollection);
                }
                /// Try to create counters
                Process process = Process.GetCurrentProcess();
                performanceCounterKeyCreationLatency = new PerformanceCounter(WCFClientCachePerformanceCounterCategoryName,
                                                                              KeyCreationLatencyPerformanceCounterName,
                                                                              string.Format("{0} ({1})", process.ProcessName, process.Id),
                                                                              false);

                // Set the raw value, this physically creates the performance counter instance
                performanceCounterKeyCreationLatency.RawValue = 0;

                // If we found one set flag to true, this will be checked later on to determine whether to increment counters
                performanceCountersExist = true; 
            }
            catch (Exception) 
            {
                // Failure usually indicates either permission problems or category does not exist, 
                // either way, we log the error and carry on without updating counters
                //ExceptionHelper.HandleException(ex);
                performanceCountersExist = false;
            }
            #endregion
        } 
        #endregion

        #region Public Static Methods
        public static string GenerateKey(Message message, OperationSettings operationSettings)
        {
            if (message == null)
            {
                throw new ArgumentException(MessageParameterCannotBeNull);
            }
            try
            {

                KeyCreationMethod keyCreationMethod = operationSettings.KeyCreationMethod == KeyCreationMethod.Indexed &&
                                                      operationSettings.Indexes.Count == 0 ?
                                                      KeyCreationMethod.Simple :
                                                      operationSettings.KeyCreationMethod;

                XmlDictionaryReader reader = null;
                StringBuilder builder = null;
                string rootName = null;
                bool ok = true;
                bool skip = false;
                int parameterDepth = 0;
                int parameterIndex = 0;
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                try
                {
                    switch (keyCreationMethod)
                    {
                        case KeyCreationMethod.Action:
                            return message.Headers.Action;

                        case KeyCreationMethod.MessageBody:
                            reader = message.GetReaderAtBodyContents();
                            return reader.ReadOuterXml();

                        case KeyCreationMethod.Simple:
                            builder = new StringBuilder();
                            builder.AppendFormat("[{0}]", message.Headers.Action);
                            reader = message.GetReaderAtBodyContents();
                            rootName = reader.Name;
                            while (reader.Read() && ok)
                            {
                                switch (reader.NodeType)
                                {
                                    case XmlNodeType.Element:
                                        builder.Append("(");
                                        break;

                                    case XmlNodeType.EndElement:
                                        if (reader.Name == rootName)
                                        {
                                            ok = false;
                                        }
                                        else
                                        {
                                            builder.Append(")");
                                        }
                                        break;

                                    case XmlNodeType.Text:
                                        builder.AppendFormat("{0}", reader.Value);
                                        break;
                                }
                            }
                            return builder.ToString();

                        case KeyCreationMethod.Indexed:
                            builder = new StringBuilder();
                            builder.AppendFormat("[{0}]", message.Headers.Action);
                            reader = message.GetReaderAtBodyContents();
                            parameterDepth = reader.Depth + 1;
                            rootName = reader.Name;
                            while (reader.Read() && ok)
                            {
                                switch (reader.NodeType)
                                {
                                    case XmlNodeType.Element:
                                        if (reader.Depth == parameterDepth)
                                        {
                                            parameterIndex++;
                                            skip = !operationSettings.Indexes.Contains(parameterIndex);
                                        }
                                        if (!skip)
                                        {
                                            builder.Append("(");
                                        }
                                        break;

                                    case XmlNodeType.EndElement:
                                        if (reader.Name == rootName)
                                        {
                                            ok = false;
                                        }
                                        else
                                        {
                                            if (!skip)
                                            {
                                                builder.Append(")");
                                            }
                                        }
                                        break;

                                    case XmlNodeType.Text:
                                        if (!skip)
                                        {
                                            builder.AppendFormat("{0}", reader.Value);
                                        }
                                        break;
                                }
                            }
                            return builder.ToString();

                        case KeyCreationMethod.MD5:
                            MD5 md5 = MD5.Create();
                            reader = message.GetReaderAtBodyContents();
                            byte[] buffer = UTF8Encoding.UTF8.GetBytes(reader.ReadOuterXml());
                            byte[] hash = md5.ComputeHash(buffer);
                            return Convert.ToBase64String(hash);

                        default:
                            reader = message.GetReaderAtBodyContents();
                            return reader.ReadOuterXml();
                    }
                }
                finally
                {
                    stopwatch.Stop();
                    if (performanceCountersExist)
                    {
                        performanceCounterKeyCreationLatency.RawValue = stopwatch.ElapsedMilliseconds;
                    }
                }
            }
            catch (Exception)
            {
                //ExceptionHelper.HandleException(ex);
                throw;
            }
        }
        #endregion
    }
}
